module Asciidoctor
  module BSI
    class Converter < ISO::Converter
      def biblio_cleanup(xmldoc)
        biblio_metanorma_ids_informative(xmldoc)
        super
      end

      # Rules for Structure 15.4
      def biblio_metanorma_ids_normative(xmldoc)
        xmldoc.xpath("//references[@normative = 'true']/bibitem")
          .each_with_index do |b, i|
          biblio_metanorma_ids1(b, i)
        end
      end

      def biblio_metanorma_ids1(bib, idx)
        bib.at("./docidentifier[@type][not(@type = 'DOI' or "\
               "@type = 'ISSN' or @type = 'ISBN')]") and return
        bib&.at("./docidentifier[@type = 'metanorma']")&.remove
        bib.at("./docidentifier").previous =
          "<docidentifier type='metanorma'>[#{idx}]</docidentifier>"
      end

      # Rules for Structure 21.4
      def biblio_metanorma_ids_informative(xmldoc)
        xmldoc.xpath("//references[@normative = 'false']/bibitem")
          .each_with_index do |b, i|
          biblio_metanorma_ids1(b, i)
        end
      end

      def biblio_reorder(xmldoc)
        xmldoc.xpath("//references[@normative = 'false']").each do |r|
          biblio_reorder1(r)
        end
        xmldoc.xpath("//references[@normative = 'true']").each do |r|
          biblio_reorder1(r)
        end
      end

      def sort_biblio(bib)
        sort_norm_appearance(bib)
      end

      def citation_order(bib)
        return {} if bib.empty?

        ids = bib.each_with_object({}) { |b, m| m[b["id"]] = nil }
        bib.first.document.xpath(".//xref").each_with_index do |b, i|
          next unless ids.has_key?(b["target"])

          ids[b["target"]] ||= i + 1
        end
        ids
      end

      # if this is a standard document, we ignore citation order
      # (Rules for Structure 21.4). That is done by zeroing id_to_num
      #
      def id_to_num(bib, ids)
        return "000000000" if bib["id"].nil?

        sprintf("%09d", ids[bib["id"]] || 0)
      end

      def sort_norm_appearance(bib)
        ids = citation_order(bib)
        bib.each do |b|
          ids[b["id"]] = 0 unless b.at("./docidentifier[@type = 'metanorma']")
        end
        bib.sort do |a, b|
          "#{id_to_num(a, ids)} :: #{sort_biblio_key(a)}" <=>
          "#{id_to_num(b, ids)} :: #{sort_biblio_key(b)}"
        end
      end

      # Rules for Structure 15.4
      # Normative references ordering
      # sort by: alphanumeric
      # prefix, then docnumber if present, numeric sort
      #      else alphanumeric metanorma id (abbreviation)
      # then doc part number if present, numeric sort
      # then doc id (not DOI &c)
      # then title
      def sort_biblio_key(bib)
        num = bib&.at("./docnumber")&.text
        id = bib&.at("./docidentifier[not(#{OTHERIDS})]")
        metaid = bib&.at("./docidentifier[@type = 'metanorma']")&.text
        abbrid = metaid unless /^\[\d+\]$/.match?(metaid)
        /\d-(?<partid>\d+)/ =~ id&.text
        type = id["type"] if id
        "#{type} :: "\
          "#{num.nil? ? abbrid : sprintf('%09d', num.to_i)} :: "\
          "#{partid} :: #{id&.text} :: #{biblio_key_title(bib)}"
      end

      def biblio_key_title(bib)
        bib&.at("./title[@type = 'main']")&.text ||
          bib&.at("./title")&.text || bib&.at("./formattedref")&.text
      end

      def biblio_renumber(xmldoc)
        ["true", "false"].each do |norm|
          i = 0
          xmldoc.xpath("//bibliography//references | //clause//references | "\
                       "//annex//references").each do |r|
            next if r["normative"] != norm

            r.xpath("./bibitem").each do |b|
              next unless docid = b.at("./docidentifier[@type = 'metanorma']")
              next unless /^\[\d+\]$/.match?(docid.text)

              i += 1
              docid.children = "[#{norm == 'true' ? 'N' : ''}#{i}]"
            end
          end
        end
      end

      # Rules for Structure 15.4
      def normref_cleanup(xmldoc)
        biblio_metanorma_ids_normative(xmldoc)
        xmldoc.xpath("//references[@normative = 'true']//references")
          .each do |ref|
          ref["unnumbered"] = true
        end
        super
      end

      def section_names_refs_cleanup(xml)
        replace_title(xml, "//bibliography/references[@normative = 'true']",
                      @i18n&.normref, true)
        replace_title(xml, "//bibliography/references[@normative = 'false']",
                      @i18n&.bibliography, true)
      end

      def new_ref(title, norm)
        id = "_#{UUIDTools::UUID.random_create}"
        "<references id='#{id}' normative='#{norm}' unnumbered='true'>"\
          "<title>#{title}</title></references>"
      end

      def boilerplate_cleanup(xmldoc)
        super
        f = xmldoc.at("//bibliography//references[@normative = 'false']"\
                      "[not(references)]")
        f and biblio_preface(f)
      end

      # Rules for Structure 21.5
      def biblio_preface(ref)
        ref.at("./title").next = "<p>#{@i18n.biblio_pref}</p>"
      end
    end
  end
end
