require_relative "./validate_requirement"
require_relative "./validate_list"

module Asciidoctor
  module BSI
    class Converter < ISO::Converter
      # BSI 0:2016 9.4.1
      def doctype_validate(xmldoc)
        doctype = xmldoc&.at("//bibdata/ext/doctype")&.text
        %w(specification management-systems-standard code-of-practice
           guide method-of-test method-of-specifying vocabulary
           classification).include? doctype or
          @log.add("Document Attributes", nil,
                   "#{doctype} is not a recognised document type")
      end

      def validate(doc)
        content_validate(doc)
        schema_validate(formattedstr_strip(doc.dup),
                        File.join(File.dirname(__FILE__), "bsi.rng"))
      end

      def image_name_validate(xmldoc); end

      def style(node, text)
        super
        style_plusminus(node, text)
        style_subscript(node)
      end

      # Rules for Structure 27.5
      def style_plusminus(node, text)
        style_regex(/\s*(?<num>\S+\u00b1\s*\S+)/,
                    "no space before plus-minus sign", node, text)
        style_regex(/\s*(?<num>\S+\s*\u00b1\s+\S+)/,
                    "space after plus-minus sign", node, text)
      end

      # Rules for Structure 27.6
      def style_subscript(node)
        node.xpath(".//sub[.//sub]").each do |r|
          style_warning(node, "Nested subscript", r)
        end
        node.xpath(".//m:msub[.//m:msub]", "m" => MATHML_NS).each do |r|
          style_warning(node, "Nested subscript", r)
        end
        node.xpath(".//sup[.//sup]").each do |r|
          style_warning(node, "Nested superscript", r)
        end
        node.xpath(".//m:msup[.//m:msup]", "m" => MATHML_NS).each do |r|
          style_warning(node, "Nested superscript", r)
        end
      end

      # protest decimal comma not decimal point
      # TODO we will not yet override decimal comma warning if this is a National Annex to Eurocode (Rules for Structure Annex B)
      def style_number(node, text)
        style_two_regex_not_prev(
          node, text, /^(?<num>-?[0-9]{4,}[,0-9]*)$/,
          %r{\b(ISO|IEC|IEEE/|(in|January|February|March|April|May|June|August|September|October|November|December)\b)$},
          "number not broken up in threes"
        )
        style_regex(/\b(?<num>[0-9]+,[0-9]+)/i,
                    "possible decimal comma", node, text)
        style_regex(/\b(?<num>billions?)\b/i,
                    "ambiguous number", node, text)
        ordinal_check(node)
      end

      # Rules for Structure 9.1
      def ordinal_check(node)
        node.xpath(".//sup").each do |s|
          next unless %w(st nd rd th).include? s.text

          style_warning(s, "text contains ordinal superscript", s.text)
        end
      end

      def section_style(root)
        super
        @doctype = root.at("//bibdata/ext/doctype")&.text
        @type = @doctype.gsub(/-/, " ").split.map(&:capitalize).join(" ")
        informative_text_style(root)
        normative_text_style(root)
        avoid_must(root)
        annex_validate(root)
      end

      def annex_validate(root)
        annex_normative_validate(root)
        annex_ordering_validate(root)
      end

      # Rules for Structure 20.2
      def annex_normative_validate(root)
        @doctype == "guide" and root.xpath("//annex[@obligation = 'normative']")
          .each do |s|
          style_warning(s, "Guide contains normative annex",
                        s&.at("./title")&.text)
        end
      end

      # Rules for Structure 20.4
      def annex_ordering_validate(root)
        ids = citation_order(root.xpath("//annex"))
        cite_order = ids.keys.sort_by { |a| ids[a] || 999999 }
        annex_order = root.xpath("//annex").map { |x| x["id"] }
        cite_order.each_with_index do |a, i|
          next if annex_order[i] == a || ids[a].nil?

          annex = root.at("//annex[@id = '#{a}']")
          title = annex&.at("./title")&.text || "Annex #{a}"
          style_warning(annex, "#{title} should be annex ##{i + 1}", nil)
        end
      end

      BSI_PUBLISHER_XPATH =
        "./contributor[role/@type = 'publisher']/"\
        "organization[abbreviation = 'ISO' or abbreviation = 'IEC' or "\
        "abbreviation = 'BSI' or abbreviation = 'CEN' or "\
        "abbreviation = 'CENELEC']".freeze

      NORM_BSI_WARN = "reference other than BSI, CEN, CENELEC, ISO, IEC "\
                      "not expected as normative".freeze

      # Rules for Structure 10.2
      def norm_bibitem_style(root)
        root.xpath(NORM_BIBITEMS).each do |b|
          b.at(BSI_PUBLISHER_XPATH).nil? and
            @log.add("Style", b, "#{NORM_BSI_WARN}: #{b.text}")
          status = b&.at("./status/stage")&.text
          /^9.$/.match?(status) and
            @log.add("Style", b,
                     "Do not cite withdrawn standards as normative: #{b.text}")
          /^[0-5].$/.match?(status) and
            @log.add("Style", b, "Do not cite unpublished standards "\
                                 "as normative: #{b.text}")
        end
      end

      # Rules for Structure 11.5.2
      def title_names_type_validate(root)
        super
        doctypes = /British\sStandard/xi
        title_main_en = title_lang_part(root, "main", "en")
        !title_main_en.nil? && doctypes.match(title_main_en.text) and
          @log.add("Style", title_main_en, "Main Title may name document type")
        title_intro_en = title_lang_part(root, "intro", "en")
        !title_intro_en.nil? && doctypes.match(title_intro_en.text) and
          @log.add("Style", title_intro_en,
                   "Title Intro may name document type")
      end

      # Rules for Structure 16.5.5
      def termdef_style(xmldoc)
        super
        xmldoc.xpath("//term").each do |t|
          para = t.at("./definition") || return
          term = t.at("./preferred").text.strip
          termdef_warn(para.text, /^[A-Z]/, t, term,
                       "term definition starts with upper case character")
          termdef_warn(para.text, /\b#{term.downcase}\b/i, t, term,
                       "term definition may be circular")
        end
      end

      def content_validate(doc)
        super
        symbol_style(doc)
        table_validate(doc)
        listcount_validate(doc)
        list_punctuation(doc)
      end

      # Rules for Structure 17.5
      def symbol_style(xmldoc)
        return if @novalid

        xmldoc.xpath("//definitions//dd").each do |para|
          term = para&.at("./preceding::dt[1]")&.text&.strip || "[symbol]"
          termdef_warn(para.text, /^(the|a)\b/i, para, term,
                       "symbol definition starts with article")
          termdef_warn(para.text, /\.$/i, para, term,
                       "symbol definition ends with period")
          termdef_warn(para.text, /^[A-Z]/, para, term,
                       "symbol definition starts with upper case character")
          termdef_warn(para.text, /\b#{term.downcase}\b/i, para, term,
                       "symbol definition may be circular")
        end
      end

      # Rules for Structure 22.3.2
      def subclause_validate(root)
        return if @novalid

        root.xpath("//clause/clause/clause/clause/clause/clause")
          .each do |c|
          style_warning(c, "Exceeds the maximum clause depth of 5", nil)
        end
      end

      # Rules for Structure 29.6
      def table_validate(xmldoc)
        xmldoc.xpath("//td[not(./*) and normalize-space(.)=''] | "\
                     "//th[not(./*) and normalize-space(.)='']").each do |td|
          style_warning(td, "Empty table cell", nil)
        end
      end

      # Generalise to all biblio: Standards Publications and Other Publications
      def normref_validate(xmldoc)
        xmldoc.xpath("//references").each do |r|
          t = r.at("./title")
          case t.text.strip.downcase
          when "standards publications"
            r.xpath("./bibitem").each { |b| normref_validate1(b, true) }
          when "other publications"
            r.xpath("./bibitem").each { |b| normref_validate1(b, false) }
          end
        end
      end

      def normref_validate1(bib, std_pub)
        numbered = bib.at("./docidentifier[@type = 'metanorma']")
        numbered && std_pub and
          style_warning(bib, "Non-Standard reference included under "\
                             "Standards Publications", nil)
        !numbered && !std_pub and
          style_warning(bib, "Standard reference included under "\
                             "Other Publications", nil)
      end

      # TODO: we will not block foreword from National Annex to Eurocode (Rules for Structure G.6.1)
      #
      # TODO: we will not test location of commentaries (Rules for Structure 24.7)
      #
      # TODO: we will not check for space either side of mathematical sign (Rules for Structure 27.5)
      #
      # TODO: we will not check for conflict of ln and log_e or log lg log_10 (Rules for Structure 27.8)
      #
      # TODO: we will not check order of brackets in stem as {[( (Rules for Structure 27.9)
    end
  end
end
