module Asciidoctor
  module BSI
    class Converter < ISO::Converter
      def external_constraint(text); end

      # Rules for Structure 7.1
      REQUIREMENT_RE_STR = "\\b(shall)\\b".freeze

      RECOMMENDATION_RE_STR = "\\b(should)\\b".freeze

      PERMISSION_RE_STR = "\\b(may)\\b".freeze

      POSSIBILITY_RE_STR = "\\b(can|might)\\b".freeze

      # Rules for Structure 7.1
      def avoid_must(root)
        root.xpath("//sections/clause[not(@type = 'scope')] | //annex")
          .each do |s|
          r = avoid_must_check(extract_text(s))
          style_warning(s, "text contains ambiguous 'must'", r) if r
        end
      end

      def avoid_must_check(text)
        text.split(/\.\s+/).each do |t|
          return t if /\b(must|will)\b/xi.match? t
        end
        nil
      end

      # Rules for Structure 7.2
      def tech_report_style(root); end

      # Rules for Structure 7.2, 7.3, 24.5
      def informative_text_style(root)
        root.xpath("//note | //admonition[@type = 'commentary'] | "\
                   "//annex[@obligation = 'informative']").each do |s|
          if r = requirement_check(extract_text(s))
            style_warning(s, "#{@type} clause may contain requirement", r)
          end
          if r = recommendation_check(extract_text(s))
            %w(code-of-practice guide).include?(@doctype) and
              style_warning(s, "#{@type} clause may contain recommendation", r)
          end
        end
      end

      # requirement language taken care of in informative_text_style
      def note_style(node)
        style(node, extract_text(node))
      end

      # Rules for Structure 25.5
      def example_style(node)
        return if @novalid

        r = requirement_check(extract_text(node)) and
          style_warning(node, "Example may contain requirement", r)
        style(node, extract_text(node))
      end

      def asset_style(root)
        root.xpath("//admonition[not(@type = 'commentary')]").each do |e|
          admonition_style(e, extract_text(e))
        end
        super
      end

      # Rules for Structure 24.8
      def admonition_style(node, text)
        r = requirement_check(text) and
          style_warning(node, "admonition may contain requirement", r)
        r = permission_check(text) and
          style_warning(node, "admonition may contain permission", r)
        r = recommendation_check(text) and
          style_warning(node, "admonition may contain recommendation", r)
        r = possibility_check(text) and
          style_warning(node, "admonition may contain possibility", r)
      end

      # Rules for Structure 7.2
      def normative_text_style(root)
        root.xpath("//sections/* | //annex[@obligation = 'normative']")
          .each do |s1|
          s = s1.dup
          s.xpath(".//note | .//admonition | .//example").each(&:remove)
          %w(code-of-practice guide).include?(@doctype) and
            r = requirement_check(extract_text(s)) and
            style_warning(s, "#{@type} clause may contain requirement", r)
          %w(specification method-of-specifying test-method)
            .include?(@doctype) and spec_normative_text_style(s)
        end
      end

      # Rules for Structure 7.3, 7.4, 7.5
      def spec_normative_text_style(node)
        r = recommendation_check(extract_text(node))
        r and style_warning(node,
                            "#{@type} clause may contain recommendation", r)
        r = permission_check(extract_text(node))
        r and style_warning(node, "#{@type} clause may contain permission", r)
        r = possibility_check(extract_text(node))
        r and style_warning(node, "#{@type} clause may contain possibility", r)
      end

      def see_xrefs_validate(root)
        root.xpath("//xref").each do |t|
          (target = root.at("//*[@id = '#{t['target']}']")) || next
          if target.at("./ancestor-or-self::annex")
            xrefs_annex_validate(t, target)
          else
            preceding = t.at("./preceding-sibling::text()[last()]")
            next unless !preceding.nil? &&
              /\b(see| refer to)\s*$/mi.match(preceding)

            if target&.at("./ancestor-or-self::*[@obligation = 'normative']")
              @log.add("Style", t, "'see #{t['target']}' is pointing "\
                       "to a normative section")
            end
          end
        end
      end

      PARENT_PARA = "./ancestor-or-self::*[./parent::clause or "\
        "./parent::annex or ./parent::terms or ./parent::definitions or "\
        "./parent::references or ./parent::*[./parent::preface]]".freeze

      # Rules for Structure 20.5
      def xrefs_annex_validate(xref, target)
        normative = !target&.at("./ancestor-or-self::*"\
                                "[@obligation='normative']").nil?
        clause = xref.at(PARENT_PARA) or return
        prec = (clause&.children & xref&.xpath("./preceding::text()"))
        &.text&.split(/\.\s*(?=[A-Z])/m)[-1]&.strip
        foll = (clause.children & xref.xpath("./following::text()"))
        &.text&.split(/\.\s*(?=[A-Z])/m)[0]&.strip
        text = "#{prec} XREF #{foll}".strip
        if %w(specification method-of-specifying).include?(@doctype) && normative
          requirement_re.match?(text) or
            style_warning(xref, "#{@type} must cite normative "\
                          "annex as requirement", text)
        elsif @doctype == "code-of-practice" && normative
          recommendation_re.match?(text) or
            style_warning(xref, "#{@type} must cite normative "\
                          "annex as recommendation", text)
        elsif %w(specification method-of-specifying).include?(@doctype) && !normative
          (recommendation_re.match?(text) ||
           (!requirement_re.match?(text) && !permission_re.match?(text) &&
            !possibility_re.match?(text))) or
          style_warning(xref, "#{@type} must cite informative annex as "\
                        "recommendation or statement", text)
        elsif %w(code-of-practice guide).include?(@doctype) && !normative
          (!recommendation_re.match?(text) && !requirement_re.match?(text) &&
           !permission_re.match?(text) && !possibility_re.match?(text)) or
          style_warning(xref, "#{@type} must cite informative annex as "\
                        "statement", text)
        end
      end

      # TODO: defer requirements and recommendations in annex (Rules for Structure 20.2)
    end
  end
end
